package de.hdm_stuttgart.mi.sd1.task2;

import de.hdm_stuttgart.mi.exam.unitmarking.ExaminationTestDefaults;
import de.hdm_stuttgart.mi.exam.unitmarking.Marking;
import de.hdm_stuttgart.mi.sd1.ignore_me.ObjectWrapper;
import org.junit.Assert;
import org.junit.FixMethodOrder;
import org.junit.Test;
import org.junit.runners.MethodSorters;
@FixMethodOrder(MethodSorters.NAME_ASCENDING)

public class Test_QuadratPolynom extends ExaminationTestDefaults {

    @Test
    @Marking(points = 5)
    public void test_100_NoZero() {
        final ObjectWrapper<QuadratPolynom> poly = new ObjectWrapper<>(QuadratPolynom.class,1, 1, 1);
        assertNoZero(poly, composeDescription(1,2,1).toString());

        final ObjectWrapper<QuadratPolynom> poly2 = new ObjectWrapper<>(QuadratPolynom.class,1, 2, 3);
        assertNoZero(1, 6, 20, poly2);
        assertNoZero(2, 13, 41, poly2);
    }

    @Test
    @Marking(points = 5)
    public void test_200_OneZero() {
        final ObjectWrapper<QuadratPolynom> poly = new ObjectWrapper<>(QuadratPolynom.class,1, 2, 1);
        assertOneZero(poly, -1, composeDescription(1,2,1).toString());

        final ObjectWrapper<QuadratPolynom> poly2 = new ObjectWrapper<>(QuadratPolynom.class,1, 2, 3);
        assertOneZero(6, -84, 294, poly2, 7);
        assertOneZero(21, 126, 189, poly2, -3);
    }

    @Test
    @Marking(points = 5)
    public void test_300_twoZeroes() {

        final ObjectWrapper<QuadratPolynom> poly = new ObjectWrapper<>(QuadratPolynom.class,1, 3, 2);
        assertTwoZeroes(poly, -2, -1, "x² + 3x + 2");

        assertTwoZeroes(3, 7, 4, poly, -4./3, -1);

        assertTwoZeroes(-2, 1, 3, poly, -1, 1.5);

        assertTwoZeroes(-1, 5, 6, poly, -1, 6);

        assertTwoZeroes(7, 11, 4, poly, -1, -4./7);

        assertTwoZeroes(4, -3, -10, poly, -1.25, 2);

        assertTwoZeroes(5, 6, 1, poly, -1, -0.2);

    }

    @Test
    @Marking(points = 2)
    public void test_400_Exception() {
        final ObjectWrapper<QuadratPolynom> poly = new ObjectWrapper<>(QuadratPolynom.class,1, -200, 2);

        final Class<ArithmeticException> expectedArithmeticException = ArithmeticException.class;

        // Coefficient ax² must not be 0
        poly.invoke(void.class, "setA", expectedArithmeticException, 0);
        new ObjectWrapper<>(QuadratPolynom.class, expectedArithmeticException, 0, 88, 99);
    }

    @Test
    @Marking(points = 3)
    public void test_500_TwoZeroExzess() {
        final ObjectWrapper<QuadratPolynom> poly = new ObjectWrapper<>(QuadratPolynom.class,
                46010, 598130,-1960026000);
        assertTwoZeroes(poly,-213,200,
                composeDescription(46010, 598130,-1960026000).toString());
    }

    // End of tests -------------------------------------------------------------------------------------------------
    // Test helper methods
    //
    static private final double delta = 1.E-12;

    static private StringBuffer composeDescription(final int a, final int b, final int c) {
        final StringBuffer ret = new StringBuffer();
        ret.append(a).append("x²");
        if (0 < b) {
            ret.append(" + ").append(b).append("x");
        } else if (b < 0) {
            ret.append(" ").append(b).append("x");
        }
        // Slight code duplication avoiding a method definition. Yepp: I'm lazy sometimes!
        if (0 < c) {
            ret.append(" + ").append(c);
        } else if (c < 0) {
            ret.append(" ").append(c);
        }
        return ret;
    }

    /**
     * Testing given polynomial for no real zero value.
     *
     * @param poly Polynomial to be tested
     * @param description E.g. "-2x² + 4x -3"
     */
    static private void assertNoZero(final ObjectWrapper<QuadratPolynom> poly, final String description) {

        Assert.assertEquals("No zero expected for polynom »" + description + "«",
                0, poly.invoke(double[].class, "getZeroes").length);
    }

    /**
     * Testing for no real valued zero solution.
     *
     * @param a Square coefficient
     * @param b Linear coefficient
     * @param c Constant coefficient
     * @param poly Square polynomial
     */
    static private void assertNoZero(final int a,
                                      final int b,
                                      final int c,
                                      final ObjectWrapper<QuadratPolynom> poly) {
        poly.invoke(void.class, "setA", a);
        poly.invoke(void.class, "setB", b);
        poly.invoke(void.class, "setC", c);
        Assert.assertEquals("One zero expected for polynom »" + composeDescription(a, b, c) + "«",
                0, poly.invoke(double[].class, "getZeroes").length);


    }

        /**
         * Testing given polynomial for one zeroes to be returned
         *
         * @param poly Polynomial to be tested
         * @param expected zero value
         * @param description E.g. "-2x² + 4x -3"
         */
    static private void assertOneZero(final ObjectWrapper<QuadratPolynom> poly,
                                      double expected,
                                      final String description) {
        final double[] result = poly.invoke(double[].class, "getZeroes");

        Assert.assertEquals("One zero expected for polynom »" + description + "«",
                1, result.length);
        Assert.assertEquals("Expected zero of »" + description +"« to be " + expected, expected, result[0], delta);
    }

    /**
     * Like {@link #assertOneZero(ObjectWrapper, double, String)} but re-setting coefficients beforehand,
     * @param a Square coefficient
     * @param b Linear coefficient
     * @param c Constant coefficient
     * @param poly Square polynomial
     */
    static private void assertOneZero(final int a,
                                      final int b,
                                      final int c,
                                      final ObjectWrapper<QuadratPolynom> poly,
                                      double expected) {
        poly.invoke(void.class, "setA", a);
        poly.invoke(void.class, "setB", b);
        poly.invoke(void.class, "setC", c);
        final double[] result = poly.invoke(double[].class, "getZeroes");

        final StringBuffer polynomDescription = composeDescription(a, b, c);

        Assert.assertEquals(2 + " zeroes expected for polynom »" + polynomDescription + "«",
                1, result.length);
        Assert.assertEquals("Expected zero of »" + polynomDescription +"« to be " + expected,
                expected, result[0], delta);
    }

    /**
     * Testing given polynomial for two zeroes to be returned
     *
     * @param poly Polynomial to be tested
     * @param expectedLower Lower zero value
     * @param expectedUpper Upper zero value
     * @param description E.g. "-2x² + 4x -3"
     */
    static private void assertTwoZeroes(final ObjectWrapper<QuadratPolynom> poly,
                                        double expectedLower,
                                        double expectedUpper,
                                        final String description) {
        final double[] result = poly.invoke(double[].class, "getZeroes");

        Assert.assertEquals(2 + " zeroes expected for polynom »" + description + "«",
                2, result.length);
        Assert.assertEquals("Expected lower zero of »" + description +"« to be " + expectedLower,
                expectedLower, result[0], delta);
        Assert.assertEquals("Expected upper zero of »" + description +"« to be " + expectedUpper,
                expectedUpper, result[1], delta);
    }

    /**
     * Like {@link #assertTwoZeroes(ObjectWrapper, double, double, String)} but re-setting coeficients beforehand,
     * @param a Square coefficient
     * @param b Linear coefficient
     * @param c Constant coefficient
     * @param poly Square polynomial
     * @param expectedLower Expected lower zero value
     * @param expectedUpper Expected upper zero value
     */
    static private void assertTwoZeroes(final int a,
                                        final int b,
                                        final int c,
                                        final ObjectWrapper<QuadratPolynom> poly,
                                        double expectedLower,
                                        double expectedUpper) {
        poly.invoke(void.class, "setA", a);
        poly.invoke(void.class, "setB", b);
        poly.invoke(void.class, "setC", c);
        final double[] result = poly.invoke(double[].class, "getZeroes");

        final StringBuffer polynomDescription = composeDescription(a, b, c);

        Assert.assertEquals(2 + " zeroes expected for polynom »" + polynomDescription + "«",
                2, result.length);
        Assert.assertEquals("Expected lower zero of »" + polynomDescription +"« to be " + expectedLower,
                expectedLower, result[0], delta);
        Assert.assertEquals("Expected upper zero of »" + polynomDescription +"« to be " + expectedUpper,
                expectedUpper, result[1], delta);
    }
}
